import random

from beat_craft_sdk.algorithms.phrase_generator_backtracking import BackTracking
from beat_craft_sdk.algorithms.scale_generator import generate_scale_with_genetic_algorithm, num_generations
from beat_craft_sdk.craft_strategy import CraftStrategy
from beat_craft_sdk.evaluation.beat_craft_evaluation import plot_fitness_over_generations, plot_pitch_diversity_over_generation, \
    plot_ga_evaluation_into_json


class CraftingBackingTrack(CraftStrategy):

    def __init__(self):
        self.pitch_fitness_per_generations = []
        self.pitch_diversity_per_generation = []

    def generate(self,output_dir,file_name, scale):
        bt = BackTracking()
        combs = bt.generate_combinations()
        flattened_phrase= random.choice(combs)
        series_scale, self.pitch_fitness_per_generations, self.pitch_diversity_per_generation = generate_scale_with_genetic_algorithm(len(flattened_phrase),scale)
        paired_notes = list(zip(flattened_phrase, series_scale))
        return paired_notes

    def evaluate(self,output_dir,file_name):
        plot_fitness_over_generations(num_generations, self.pitch_fitness_per_generations,output_dir,file_name)
        plot_pitch_diversity_over_generation(num_generations, self.pitch_diversity_per_generation,output_dir,file_name)
        plot_ga_evaluation_into_json(self.pitch_diversity_per_generation,self.pitch_fitness_per_generations,output_dir,file_name)


import os
import  platform

from midi2audio import FluidSynth


def get_default_sound_font():
    current_os = platform.system()
    if current_os=='Linux':
        return '/usr/share/sounds/sf2/FluidR3_GM.sf2'
    elif current_os=='Darwin':
        return '/Library/Audio/Sounds/Banks/FluidR3_GM.sf2'
    else:
        raise Exception(f"Unsupported OS:{current_os}. Please provide a valid sound font path.")


class AudioConverter:
    def __init__(self,input_path,output_path, sound_font=None):
        self._input_path = input_path
        self._output_path = output_path
        self.sound_font = sound_font or os.getenv('SOUND_FONT_PATH') or get_default_sound_font()

    def midi_to_wav(self):
        try:
            if not os.path.exists(self._input_path):
                raise FileNotFoundError(f"Input MIDI file not found : {self._input_path}")
            if not os.path.exists(self.sound_font):
                raise FileNotFoundError(f"SoundFont file not found : {self.sound_font}")

            fs = FluidSynth(self.sound_font)
            fs.midi_to_audio(self._input_path,self._output_path)
        except Exception as e:
            print(f"An error occurred during conversion: {e}")


import unittest

from beat_craft_sdk.algorithms.phrase_generator_backtracking import BackTracking
from beat_craft_sdk.algorithms.phrase_generator_genetic import generate_phrase_with_genetic_algorithm


class TestBeatCraftSdk(unittest.TestCase):
    def test_generate_population(self):
        series_duration=generate_phrase_with_genetic_algorithm()
        print(series_duration)

    def test_backingtrack(self):
        backingtrack = BackTracking()
        combs = backingtrack.generate_combinations()
        for co in combs:
            print(f" sum {sum(co)} in {co}")


import unittest

from beat_craft_sdk import Config

class TestConfig(unittest.TestCase):
    def test_default_config(self):
        config = Config()
        self.assertEqual(config.tempo,"medium")
        self.assertEqual(config.vibe,"neutral")

    def test_custom_config(self):
        config = Config(tempo="fast", vibe="calm")
        self.assertEqual(config.tempo,"fast")
        self.assertEqual(config.vibe,"calm")
    def test_invalid_config(self):
        with self.assertRaises(ValueError):
            Config(tempo="extreme",vibe="neutral")

if __name__ == '__main__':
    unittest.main()


import torchaudio
from audiocraft.models import MusicGen
from audiocraft.data.audio import audio_write
import os

output_dir = ".output"
model = MusicGen.get_pretrained('melody')
model.set_generation_params(duration=8)  # generate 8 seconds.

descriptions = ['happy rock', 'energetic EDM', 'sad jazz']

melody, sr = torchaudio.load('./.output/output.wav')
# generates using the melody from the given audio and the provided descriptions.
wav = model.generate_with_chroma(descriptions, melody[None].expand(3, -1, -1), sr)


# Create the directory if it doesn't exist
os.makedirs(output_dir, exist_ok=True)

for idx, one_wav in enumerate(wav):
    output_filename = f'{idx}'
    # Full path where the audio will be saved
    output_path = os.path.join(output_dir, output_filename)
    # Will save under {idx}.wav, with loudness normalization at -14 db LUFS.
    audio_write(stem_name=output_path, wav=one_wav.cpu(), sample_rate=model.sample_rate, strategy="loudness")



import matplotlib.pyplot as plt
import os
from beat_craft_sdk.utils.beat_craft_utils import get_current_time
import numpy as np
import librosa
import json

def fitness_smoothness(sequence):
    total_jump = sum(abs(sequence[i] - sequence[i + 1]) for i in range(len(sequence) - 1))
    return 1 / (1 + total_jump)

consonant_intervals = {0, 7, 12}
def fitness_consonance(sequence):
    consonant_pairs = sum(1 for i in range(len(sequence) - 1) if abs(sequence[i] - sequence[i + 1]) in consonant_intervals)
    return consonant_pairs / (len(sequence) - 1)

def fitness_rhythmic_variety(sequence):
    variety = len(set(sequence)) / len(sequence)
    return variety

def combined_fitness(sequence):
    smoothness_score = fitness_smoothness(sequence)
    consonance_score = fitness_consonance(sequence)
    variety_score = fitness_rhythmic_variety(sequence)
    return 0.4 * smoothness_score + 0.4 * consonance_score + 0.2 * variety_score

# Hamming distance to measure genotypic diversity
def hamming_distance(seq1, seq2):
    return sum(el1 != el2 for el1, el2 in zip(seq1, seq2))

def population_diversity(population):
    total_distance = 0
    count = 0
    for i in range(len(population)):
        for j in range(i + 1, len(population)):
            total_distance += hamming_distance(population[i], population[j])
            count += 1
    if count == 0:
        return 0
    return total_distance / count  # Return average Hamming distance

# Fitness diversity (standard deviation of fitness scores)
def fitness_diversity(fitness_scores):
    return np.std(fitness_scores)

# Count the number of unique individuals in the population
def unique_individuals(population):
    unique_sequences = set(tuple(seq) for seq in population)
    return len(unique_sequences) / len(population)  # Fraction of unique sequences

def plot_fitness_over_generations(num_generation,fitness_per_generations,output_dir,file_name):
    filename = f"fitness_score_evaluation_{get_current_time()}_{file_name}.png"
    output_path = os.path.join(output_dir, filename)
    plt.figure()
    plt.plot(range(1,num_generation+1),fitness_per_generations,marker='o')
    plt.title('Scale-Fitness Score Evaluation Over Generations')
    plt.xlabel('Generation')
    plt.ylabel('Best Fitness Score')
    plt.grid(True)
    plt.savefig(output_path)

def plot_pitch_diversity_over_generation(num_generation,diversity_per_generation,output_dir,file_name):
    filename = f"pitch_diversity_evaluation_{get_current_time()}_{file_name}.png"
    output_path = os.path.join(output_dir, filename)
    plt.figure()
    plt.plot(range(1, num_generation + 1), diversity_per_generation, marker='o', color='orange')
    plt.title('Pitch Diversity Evolution Over Generations')
    plt.xlabel('Generation')
    plt.ylabel('Pitch Population Diversity (Hamming Distance)')
    plt.grid(True)
    plt.savefig(output_path)

def plot_waveform(audio_data,sample_rate, output_dir,file_name):
    validate_audio_data(audio_data)
    filename = f"waveform_generated_music_{get_current_time()}_{file_name}.png"
    output_path = os.path.join(output_dir, filename)
    # Create time axis
    time = np.linspace(0, len(audio_data) / sample_rate, num=len(audio_data))
    # Plot the waveform
    plt.figure(figsize=(10, 4))
    plt.plot(time, audio_data)
    plt.title("Waveform of Generated Music")
    plt.xlabel("Time (seconds)")
    plt.ylabel("Amplitude")
    plt.savefig(output_path)

    plot_audio_analyze_into_json('Time',
                                 time,
                                 'Amplitude',
                                 audio_data,
                                 output_dir,
                                 file_name)

def plot_spectrogram(audio_data,sample_rate, output_dir,file_name):
    validate_audio_data(audio_data)
    filename = f"spectrogram_generated_music_{get_current_time()}_{file_name}.png"
    output_path = os.path.join(output_dir, filename)
    plt.figure(figsize=(10, 6))
    D = librosa.amplitude_to_db(np.abs(librosa.stft(audio_data)), ref=np.max)
    librosa.display.specshow(D, sr=sample_rate, x_axis='time', y_axis='log')
    plt.colorbar(format='%+2.0f dB')
    plt.title('Spectrogram of Generated Music')
    plt.savefig(output_path)

    # Extract X data (time) from the frames
    time = librosa.frames_to_time(np.arange(D.shape[1]), sr=sample_rate)

    # Extract Y data (frequency in Hz)
    frequencies = librosa.fft_frequencies(sr=sample_rate)

    plot_audio_analyze_into_json('Time',
                                 time,
                                 'Spectrogram (log)',
                                 frequencies,
                                 output_dir,
                                 filename)

def plot_mel_spectrogram(audio_data,sample_rate, output_dir,file_name):
    validate_audio_data(audio_data)
    filename = f"mel_spectrogram_generated_music_{get_current_time()}_{file_name}.png"
    output_path = os.path.join(output_dir, filename)
    mel_spect = librosa.feature.melspectrogram(y=audio_data, sr=sample_rate, n_mels=128)
    mel_spect_db = librosa.power_to_db(mel_spect, ref=np.max)
    # Check the shape of mel_spectrogram_db to ensure it's in the correct format
    print("Mel Spectrogram shape1:", mel_spect_db.shape)  # Should be (128, n_time_frames)
    mel_spect_db = np.squeeze(mel_spect_db)
    print("Mel Spectrogram shape2:", mel_spect_db.shape)

    # Compute the Short-Time Fourier Transform (STFT)
    stft_result = librosa.stft(audio_data)
    D = librosa.amplitude_to_db(np.abs(stft_result), ref=np.max)

    # Extract X data (time) from the frames
    time = librosa.frames_to_time(np.arange(D.shape[1]), sr=sample_rate)

    # Extract Y data (frequency in Hz)
    frequencies = librosa.fft_frequencies(sr=sample_rate)  # For linear frequency scale

    # Plot the mel spectrogram
    plt.figure(figsize=(10, 6))
    librosa.display.specshow(D, sr=sample_rate, x_axis='time', y_axis='mel')
    plt.colorbar(format='%+2.0f dB')
    plt.title('Mel Spectrogram of Generated Music')
    plt.savefig(output_path)
    plot_audio_analyze_into_json('Time',
                                 time,
                                 'Mel',
                                 frequencies,
                                 output_dir,
                                 filename)

def plot_pitch_contour(audio_data,sample_rate, output_dir,file_name):
    validate_audio_data(audio_data)
    filename = f"pitch_contour_generated_music_{get_current_time()}_{file_name}.png"
    output_path = os.path.join(output_dir, filename)
    # Pitch and harmonic analysis
    harmonic, _ = librosa.effects.hpss(audio_data)
    pitch = librosa.yin(harmonic, fmin=librosa.note_to_hz('C2'), fmax=librosa.note_to_hz('C7'))

    time = librosa.frames_to_time(np.arange(len(pitch)), sr=sample_rate)
    # Plot pitch contour
    plt.figure(figsize=(10, 4))
    plt.plot(pitch, label='Pitch Contour')
    plt.title('Pitch Contour of Generated Music')
    plt.xlabel('Time (frames)')
    plt.ylabel('Pitch (Hz)')
    plt.legend()
    plt.savefig(output_path)
    plot_audio_analyze_into_json('Time (frames)',
                                 time,
                                 'Pitch (Hz)',
                                 pitch,
                                 output_dir,
                                 filename)
def validate_audio_data(audio_data):
    if len(audio_data) == 0 or np.all(audio_data == 0):
        print("Error: Audio data is empty or consists of only zeros.")
    else:
        print("Audio data is valid.")

def load_audio_file(file_path):
    # Load the audio file with librosa
    try:
        audio_data, sample_rate = librosa.load(file_path, sr=None)  # sr=None to preserve the original sample rate
        print(f"Loaded audio file: {file_path}")
        print(f"Sample rate: {sample_rate}")
        print(f"Audio data shape: {audio_data.shape}")
        return audio_data, sample_rate
    except Exception as e:
        print(f"Error loading audio file: {e}")
        return None, None

def plot_ga_evaluation_into_json(diversity_per_generation, best_fitness_per_generation,output_dir,file_name):
    filename = f"ga_evaluation_{get_current_time()}_{file_name}.json"
    output_path = os.path.join(output_dir, filename)
    data = {
        "diversity_per_generation": diversity_per_generation,
        "best_fitness_per_generation": best_fitness_per_generation
    }
    # Save to a JSON file
    with open(output_path, 'w') as json_file:
        json.dump(data, json_file, indent=4)

def plot_audio_analyze_into_json(x_label,x_data,y_label,y_data,output_dir,file_name):
    filename = f"audio_analyze_{get_current_time()}_{file_name}.json"
    output_path = os.path.join(output_dir, filename)
    data = {
        f"{x_label}": json.dumps(x_data.tolist()),
        f"{y_label}": json.dumps(y_data.tolist())
    }
    # Save to a JSON file
    with open(output_path, 'w') as json_file:
        json.dump(data, json_file, indent=4)


from beat_craft_sdk.config import BeatCraftConfig
from beat_craft_sdk.crafting_with_genetic import CraftingGenetic
import mido
from mido import MidiFile, MidiTrack, Message
import pygame
import os
import torchaudio
from audiocraft.models import MusicGen
from audiocraft.data.audio import audio_write

from beat_craft_sdk.evaluation.beat_craft_evaluation import plot_mel_spectrogram, plot_waveform, plot_spectrogram, plot_pitch_contour, \
    load_audio_file
from beat_craft_sdk.utils.beat_craft_utils import get_current_time


class BeatCraft:
    def __init__(self, config=None, melody_engine=None):
        self.config = config if config else BeatCraftConfig()
        self.melody_engine = melody_engine if melody_engine else CraftingGenetic()

    def set_melody_engine(self, melody_engine):
        self.melody_engine = melody_engine

    def compose_melody(self):
        notes = self.melody_engine.generate(self.config.get_output_dir(),self.config.get_file_name(), self.config.get_midi_notes(self.config.get_game_emotional()))
        self.melody_engine.evaluate(self.config.get_output_dir(),self.config.get_file_name())
        print(f"notes in core generate music {notes}")
        return notes

    def melody_to_midi(self, generated_notes=None):
        mid = MidiFile()
        track = MidiTrack()
        mid.tracks.append(track)

        # Set a tempo (BPM)
        tempo = mido.bpm2tempo(120)  # Convert BPM to MIDI tempo
        track.append(mido.MetaMessage('set_tempo', tempo=tempo))

        # Define a base time unit for quarter notes (in MIDI ticks, typically 480 ticks per beat)
        ticks_per_beat = 480

        for duration, pitch in generated_notes:
            # Convert the duration (in beats) to MIDI ticks
            ticks = int(duration * ticks_per_beat)

            if pitch != 0:  # 0 means rest, so skip note_on/note_off
                # Add a note on event (velocity 64 for normal sound)
                track.append(Message('note_on', note=pitch, velocity=64, time=0))

                # Add a note off event after the note duration
                track.append(Message('note_off', note=pitch, velocity=64, time=ticks))
            else:
                # If it's a rest, just wait (no note_on or note_off)
                track.append(Message('note_off', note=0, velocity=0, time=ticks))

        # Save the MIDI file
        output_path = f"{self.config.get_output_dir()}/{self.config.get_file_name()}.mid"
        print(f"output_path {output_path} ")
        mid.save(output_path)

    def play_generated_music(self, path):
        pygame.mixer.init()
        pygame.mixer.music.load(path)
        pygame.mixer.music.play()

        while pygame.mixer.music.get_busy():
            pass
    def generate_rythm(self,melody_file_name):

        audio_melody_path = f"{self.config.get_output_dir()}/{melody_file_name}.wav"
        model = MusicGen.get_pretrained('melody')
        model.set_generation_params(duration=8)  # generate 8 seconds.
        desc = f"{self.config.get_game_emotional()} {self.config.get_game_mood()}"
        descriptions = [desc]

        melody, sr = torchaudio.load(audio_melody_path)
        # generates using the melody from the given audio and the provided descriptions.
        wav = model.generate_with_chroma(descriptions, melody[None].expand(1, -1, -1), sr)

        for idx, one_wav in enumerate(wav):
            output_filename = f'{idx}_{get_current_time()}_{self.config.get_file_name()}'
            # Full path where the audio will be saved
            output_path = os.path.join(self.config.get_output_dir(), output_filename)
            # Will save under {idx}.wav, with loudness normalization at -14 db LUFS.
            audio_write(stem_name=output_path, wav=one_wav.cpu(), sample_rate=model.sample_rate, strategy="loudness")

            audio_data, sample_rate = load_audio_file(f"{output_path}.wav")

            plot_pitch_contour(audio_data,sample_rate, self.config.get_output_dir(), self.config.get_file_name())
            plot_mel_spectrogram(audio_data,sample_rate,self.config.get_output_dir(), self.config.get_file_name())
            plot_waveform(audio_data, sample_rate, self.config.get_output_dir(), self.config.get_file_name())
            plot_spectrogram(audio_data, sample_rate, self.config.get_output_dir(), self.config.get_file_name())

    def get_config(self):
        return self.config


import random

from beat_craft_sdk.evaluation.beat_craft_evaluation import population_diversity, fitness_consonance

# Define the A minor scale using MIDI numbers
# A_minor_scale_midi = [0, 57, 59, 60, 62, 64, 65, 67]  # A, B, C, D, E, F, G

# User input for series length and population size
series_length = 4
population_size = 100
num_generations = 500

# Initialize population
def initialize_population(population_size, series_length, scale):
    population = []
    for _ in range(population_size):
        sequence = [random.choice(scale) for _ in range(series_length)]
        population.append(sequence)
    return population


# Selection function: Selects the top sequences based on fitness
def selection(population, fitness_scores, num_parents):
    selected = [population[i] for i in
                sorted(range(len(fitness_scores)), key=lambda k: fitness_scores[k], reverse=True)[:num_parents]]
    return selected


# Crossover function: Combines two parent sequences
def crossover(parent1, parent2):
    crossover_point = random.randint(1, len(parent1) - 2)
    child1 = parent1[:crossover_point] + parent2[crossover_point:]
    child2 = parent2[:crossover_point] + parent1[crossover_point:]
    return child1, child2


# Mutation function: Mutates a sequence
def mutate(sequence, mutation_rate=0.1,scale=None):
    for i in range(len(sequence)):
        if random.random() < mutation_rate:
            sequence[i] = random.choice(scale)
    return sequence


# Main genetic algorithm
def generate_scale_with_genetic_algorithm(num_item=4, scale=None):
    population = initialize_population(population_size, num_item, scale)

    best_fitness_per_generation = []  # To store the best fitness of each generation
    diversity_per_generation = []     # To store the diversity of each generation (genotypic)

    for generation in range(num_generations):
        # Evaluate fitness of each sequence
        fitness_scores = [fitness_consonance(sequence) for sequence in population]

        best_fitness = max(fitness_scores)
        best_fitness_per_generation.append(best_fitness)

        # Measure diversity of the population
        diversity = population_diversity(population)
        diversity_per_generation.append(diversity)

        parents = selection(population, fitness_scores, population_size // 2)

        # Generate the next generation
        next_generation = []
        while len(next_generation) < population_size:
            parent1, parent2 = random.sample(parents, 2)
            child1, child2 = crossover(parent1, parent2)
            next_generation.append(mutate(sequence=child1,scale=scale))
            next_generation.append(mutate(sequence=child2,scale=scale))

        # Replace old population with new generation
        population = next_generation[:population_size]

        # Print progress
        # print(f"Generation {generation + 1}: Best fitness: {best_fitness}, Diversity: {diversity}")

    # Return the best sequence
    best_sequence = max(population, key=lambda seq: fitness_consonance(seq))
    return best_sequence, best_fitness_per_generation, diversity_per_generation


# Run the genetic algorithm and print the best sequence
# best_sequence = scale_genetic_algorithm()
# print("Best sequence found (MIDI numbers):", best_sequence)



import os.path
import unittest

from beat_craft_sdk.config import BeatCraftConfig
from beat_craft_sdk.core import BeatCraft
from beat_craft_sdk.crafting_with_backingtrack import CraftingBackingTrack
from beat_craft_sdk.utils.audio_converter import AudioConverter
from beat_craft_sdk.config import GameMood, GameEmotional, GameGenre

class TestBeatCraftSdk(unittest.TestCase):

    def test_config_all_params_none(self):
        conf = BeatCraftConfig()
        self.assertEqual(conf.get_output_dir(),"./../.outputx")
        self.assertTrue(conf.get_file_name())

    def test_config_filled_all_params(self):
        conf = BeatCraftConfig("./../.outputz","malam")
        self.assertEqual(conf.get_output_dir(), "./../.outputz")
        self.assertEqual(conf.get_file_name(),"malam")

    def test_generate_melody_with_genetic(self):
        config = BeatCraftConfig("./../.outputf","merbabu")
        sdk = BeatCraft(config)
        notes = sdk.compose_melody()
        print(f"notes {notes}")
        sdk.melody_to_midi(notes)
        self.assertTrue(os.path.exists(f"{config.get_output_dir()}/{config.get_file_name()}.mid"))

    def test_config_input_parameter(self):
        conf = BeatCraftConfig()
        self.assertEqual(GameEmotional.EXCITEMENT.value, conf.get_game_emotional())  # The default emotional is excited
        self.assertEqual(GameMood.JOYFUL.value, conf.get_game_mood())  # The default mood is HAPPY
        self.assertEqual(GameGenre.ACTION.value, conf.get_game_genre())  # the default genre is action

        conf.set_game_mood(GameMood.EPIC)
        self.assertEqual(GameMood.EPIC.value, conf.get_game_mood())

        conf.set_game_genre(GameGenre.RPG)
        self.assertEqual(GameGenre.RPG.value, conf.get_game_genre())

        conf.set_game_emotional(GameEmotional.FEAR)
        self.assertEqual(GameEmotional.FEAR.value, conf.get_game_emotional())

    def test_scale(self):
        conf = BeatCraftConfig()
        emotion = conf.get_game_emotional()
        midi_notes = conf.get_midi_notes(emotion)
        self.assertEqual([60, 62, 64, 65, 67, 69, 71],midi_notes)  #The default is C major [60, 62, 64, 65, 67, 69, 71]

    def test_generate_melody_with_backingtrack(self):
        btconfig = BeatCraftConfig(file_name='output_bt')
        sdk = BeatCraft(btconfig)
        sdk.set_melody_engine(CraftingBackingTrack())
        notes = sdk.compose_melody()
        self.assertGreater(len(notes),0,"list of notes are empty")

        sdk.melody_to_midi(notes)
        self.assertTrue(os.path.exists(f"{btconfig.get_output_dir()}/{btconfig.get_file_name()}.mid"))
    def test_generate_melody_controlled_input(self):
        btconfig = BeatCraftConfig(output_dir="./../.outputr", file_name='semut3')
        sdk = BeatCraft(btconfig)

        btconfig.set_game_mood(GameMood.JOYFUL)
        btconfig.set_game_genre(GameGenre.ACTION)
        btconfig.set_game_emotional(GameEmotional.EXCITEMENT)

        notes = sdk.compose_melody()
        sdk.melody_to_midi(notes)
        sdk.play_generated_music(f"{btconfig.get_output_dir()}/{btconfig.get_file_name()}.mid")

        conv = AudioConverter(f"{btconfig.get_output_dir()}/{btconfig.get_file_name()}.mid",
                              f"{btconfig.get_output_dir()}/{btconfig.get_file_name()}.wav")
        conv.midi_to_wav()


        sdk.generate_rythm(sdk.get_config().get_file_name())

        pass

    def test_sdk_play_midi_generated(self):
        config = BeatCraftConfig(output_dir="../.outputf", file_name='merbabu')
        sdk = BeatCraft(config)
        sdk.play_generated_music(f"{config.get_output_dir()}/{config.get_file_name()}.mid")

    def test_convert_midi_to_wav(self):
        config = BeatCraftConfig(file_name='output_bt')
        conv = AudioConverter(f"{config.get_output_dir()}/{config.get_file_name()}.mid",
                              f"{config.get_output_dir()}/{config.get_file_name()}.wav")
        conv.midi_to_wav()

    def test_generate_rythm(self):
        config = BeatCraftConfig(file_name='output_bt')
        sdk = BeatCraft(config)
        sdk.generate_rythm(sdk.get_config().get_file_name())

    def test_genetic_fitness_over_generation(self):
        pass
if __name__ == '__main__':
    unittest.main()


from abc import abstractmethod


class CraftStrategy():
    @abstractmethod
    def generate(self,output_dir,file_name, scale):
        pass

    @abstractmethod
    def evaluate(self,output_dir,file_name):
        pass




class BackTracking:
    def __init__(self, max_sum=4, max_items=16):
        self.valid_combinations = []
        # Define the note durations
        self.note_durations = {
            "whole": 4,
            "half": 2,
            "quarter": 1,
            "eighth": 0.5,
            "sixteenth": 0.25
        }

        # Maximum sum and maximum number of items
        self.max_sum = max_sum
        self.max_items = max_items

    # Backtracking function to generate all valid combinations
    def find_combinations(self, current_combo, current_sum, note_count):
        # If the current sum exceeds the maximum or note count exceeds the limit, return
        if current_sum > self.max_sum  or note_count > self.max_items:
            return

        # If the current sum is exactly equal to max_sum, store the combination
        if current_sum == self.max_sum:
            # print(f"current sum {current_sum} - note count {note_count}")
            self.valid_combinations.append(current_combo)
            return

        # Recursively add more notes and explore further combinations
        for note, duration in self.note_durations.items():
            self.find_combinations(current_combo + [duration], current_sum + duration, note_count + 1)

    def generate_combinations(self):
        self.find_combinations(current_combo=[], current_sum=0, note_count=0)
        return self.valid_combinations

# Usage example
# generator = BackTracking()
# generator.generate_combinations()



from phrase_generator_genetic import generate_phrase_with_genetic_algorithm
from scale_generator import generate_scale_with_genetic_algorithm
import mido
from mido import MidiFile, MidiTrack, Message

solution = generate_phrase_with_genetic_algorithm()
flattened_list = [item for sublist in solution for item in (sublist if isinstance(sublist, list) else [sublist])]
print(f"flattened list {flattened_list}")

numberitem = len(flattened_list)
best_sequence = generate_scale_with_genetic_algorithm(numberitem)
print(f"best_sequence list {best_sequence}")

paired_notes = list(zip(flattened_list,best_sequence))

print(paired_notes)

# Create a new MIDI file and track
mid = MidiFile()
track = MidiTrack()
mid.tracks.append(track)

# Set a tempo (BPM)
tempo = mido.bpm2tempo(120)  # Convert BPM to MIDI tempo
track.append(mido.MetaMessage('set_tempo', tempo=tempo))

# Define a base time unit for quarter notes (in MIDI ticks, typically 480 ticks per beat)
ticks_per_beat = 480

# Add notes to the MIDI track
for duration, pitch in paired_notes:
    # Convert the duration (in beats) to MIDI ticks
    ticks = int(duration * ticks_per_beat)

    if pitch != 0:  # 0 means rest, so skip note_on/note_off
        # Add a note on event (velocity 64 for normal sound)
        track.append(Message('note_on', note=pitch, velocity=64, time=0))

        # Add a note off event after the note duration
        track.append(Message('note_off', note=pitch, velocity=64, time=ticks))
    else:
        # If it's a rest, just wait (no note_on or note_off)
        track.append(Message('note_off', note=0, velocity=0, time=ticks))

# Save the MIDI file
mid.save('output.mid')

print("MIDI file 'output.mid' has been created!")


import os
from enum import Enum

from beat_craft_sdk.utils.beat_craft_utils import get_current_time

class GameGenre(Enum):
    ACTION="Energetic"
    RPG = "Epic"
    Horror = "Dissonant"
    Puzzle = "Ambient"

class GameMood(Enum):
    JOYFUL="Upbeat And Lively"
    SERENE="Calm and Ambient"
    EPIC="Grand and orchestral"
    RELAXING="Relaxing"
    TENSE="Dark and suspenseful"

class GameEmotional(Enum):#mapping into music scale
    EXCITEMENT = "Major"
    FEAR = "Minor"
    PEACEFUL = "Pentatonic"
    CHALLENGE = "Dorian"


class BeatCraftConfig:

    DEFAULT_OUTPUT_DIR = './../.outputx'
    C_major = [60, 62, 64, 65, 67, 69, 71]
    C_minor = [60, 62, 63, 65, 67, 68, 70]
    C_pentatonic = [60, 62, 64, 67, 69]
    C_dorian = [60, 62, 63, 65, 67, 69, 70]
    emotional_to_midi = {
        GameEmotional.EXCITEMENT.value: C_major,
        GameEmotional.FEAR.value: C_minor,
        GameEmotional.PEACEFUL.value: C_pentatonic,
        GameEmotional.CHALLENGE.value: C_dorian
    }

    def __init__(self,output_dir=None, file_name=None):
        self.output_dir = output_dir
        self.file_name = file_name
        self.path_midi_file = ''
        self.path_wav_file = ''
        self.genre = GameGenre.ACTION.value
        self.mood = GameMood.JOYFUL.value
        self.emotional = GameEmotional.EXCITEMENT.value
        self.validate_config()
        print(f" input default emot:{self.emotional}, genre:{self.genre}, mood:{self.mood}")

    def validate_config(self):
        # If output_dir is None or doesn't exist, create it
        if self.output_dir is None:
            os.makedirs(self.DEFAULT_OUTPUT_DIR, exist_ok=True)
            self.output_dir = self.DEFAULT_OUTPUT_DIR
            # self.output_dir = os.path.abspath('./../.outputx')  # Using absolute path for consistency
        elif not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir, exist_ok=True)  # Create the directory if it doesn't exist

        if self.file_name is None:
            self.file_name = get_current_time()

        print(f"validating config output_dir {self.output_dir}")
        print(f"validating config default_file_name {self.file_name}")

    def get_output_dir(self):
        return self.output_dir

    def get_file_name(self):
        return self.file_name

    def set_game_genre(self,game_genre:GameGenre):
        self.genre = game_genre.value

    def get_game_genre(self):
        return self.genre

    def set_game_mood(self,game_mood:GameMood):
        self.mood=game_mood.value

    def get_game_mood(self):
        return self.mood

    def set_game_emotional(self,game_emotional:GameEmotional):
        self.emotional = game_emotional.value

    def get_game_emotional(self):
        return self.emotional

    def get_midi_notes(self, game_emotional):
        return self.emotional_to_midi.get(game_emotional, "Unknown Emotion")


import random

# Define the possible note durations (in beats) in a 4/4 time signature
note_durations = {
    "whole": 4,
    "half": 2,
    "quarter": 1,
    "eighth": 0.5,
    "sixteenth": 0.25
}

# Total number of beats for 8 bars in 4/4 time signature
total_beats = 16
beats_per_bar = 4

# List of possible note durations
duration_values = list(note_durations.values())

# Function to create an initial population (randomly generate sequences of notes)
def create_individual():
    sequence = []
    total = 0
    while total < total_beats:
        note = random.choice(duration_values)
        if total + note <= total_beats:
            sequence.append(note)
            total += note
    return sequence


# Grouping function to ensure notes are grouped by bars of exactly 4 beats
def group_by_bar(note_sequence, bar_length=4):
    grouped_bars = []
    current_bar = []
    total_in_bar = 0

    for note in note_sequence:
        if total_in_bar + note <= bar_length:
            current_bar.append(note)
            total_in_bar += note
        else:
            # If adding the note exceeds 4 beats, start a new bar
            grouped_bars.append(current_bar)
            current_bar = [note]
            total_in_bar = note

        # If the bar sums exactly to 4, finalize the bar
        if total_in_bar == bar_length:
            grouped_bars.append(current_bar)
            current_bar = []
            total_in_bar = 0

    # If there are remaining notes, add them as the last bar
    if current_bar:
        grouped_bars.append(current_bar)

    return grouped_bars


# Fitness function: Penalize individuals that don't fit perfectly into 8 bars of 4 beats each
def fitness(individual):
    grouped_bars = group_by_bar(individual)

    # Check if the total number of bars is 8 and each bar sums exactly to 4 beats
    if len(grouped_bars) == 8 and all(sum(bar) == beats_per_bar for bar in grouped_bars):
        return 0  # Perfect fit
    else:
        # Penalize based on the deviation from the ideal
        return abs(total_beats - sum([sum(bar) for bar in grouped_bars])) + len(grouped_bars) - 8


# Selection function (select individuals based on fitness)
def selection(population):
    population.sort(key=lambda x: fitness(x))
    return population[:len(population) // 2]


# Crossover function (combine two individuals to create offspring)
def crossover(parent1, parent2):
    point = random.randint(1, min(len(parent1), len(parent2)) - 1)
    child1 = parent1[:point] + parent2[point:]
    child2 = parent2[:point] + parent1[point:]
    return child1, child2


# Mutation function (randomly change one note duration in the individual)
def mutation(individual):
    if random.random() < 0.1:  # 10% chance of mutation
        idx = random.randint(0, len(individual) - 1)
        individual[idx] = random.choice(duration_values)
    return individual


# Genetic Algorithm function
def generate_phrase_with_genetic_algorithm(generations=1000, population_size=50):
    population = [create_individual() for _ in range(population_size)]
    for generation in range(generations):
        # Selection
        population = selection(population)

        # Crossover
        offspring = []
        for i in range(0, len(population), 2):
            if i + 1 < len(population):
                child1, child2 = crossover(population[i], population[i + 1])
                offspring.append(child1)
                offspring.append(child2)

        # Mutation
        population = [mutation(ind) for ind in offspring]

        # Add new random individuals to maintain diversity
        while len(population) < population_size:
            population.append(create_individual())

        # Check if we found a perfect solution
        best_individual = min(population, key=lambda x: fitness(x))
        # print(f"fitness {best_individual} found in generation {generation + 1}")
        # print(f"fitness {fitness(best_individual)} found in generation {generation + 1}")
        if fitness(best_individual) == 0:
            print(f"Solution found in generation {generation + 1}")
            return best_individual

    # Return the best individual after all generations
    return min(population, key=lambda x: fitness(x))


# Running the genetic algorithm
# solution = genetic_algorithm()
# grouped_solution = group_by_bar(solution)
#
# # Output the generated solution
# print(f"Generated Note Durations for 8 Bars:{grouped_solution}")
# for bar in grouped_solution:
#     print(bar)
# print("Total Duration in each bar (should be exactly 4):", [sum(bar) for bar in grouped_solution])



from setuptools import setup, find_packages

setup(name="beatcraft",
      version="1.1.4",
      author="Arul and friends",
      author_email="arif.akbarul@amikom.ac.id",
      description="while you are focus on the game logic, "
                  "BeatCraft help you to make an authentic music for your game",
      url="",
      packages=find_packages(),
      install_requires=[
        'click',
        'MIDIUtil',
        'pyo',
        'pygame',
        'numpy==1.26.4',
        'sounddevice',
        'mido',
        'python-rtmidi',
        'pretty_midi',
        'torch==2.1.0',
        'torchaudio==2.1.0',
        'torchtext==0.16.0',
        'torchvision==0.16.0',
        'transformers==4.45.1',
        'xformers==0.0.22.post7',
        'audiocraft==1.3.0',
        'midi2audio',
        'librosa'
      ],
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent'
      ],
      python_requires=">=3.6")


from random import choices, randint, randrange, random, sample
from typing import List, Optional, Callable, Tuple

Genome = List[int]
Population = List[Genome]
PopulateFunc = Callable[[], Population]
FitnessFunc = Callable[[Genome], int]
SelectionFunc = Callable[[Population, FitnessFunc], Tuple[Genome, Genome]]
CrossoverFunc = Callable[[Genome, Genome], Tuple[Genome, Genome]]
MutationFunc = Callable[[Genome], Genome]
PrinterFunc = Callable[[Population, int, FitnessFunc], None]


def generate_genome(length: int) -> Genome:
    return choices([0, 1], k=length)


def generate_population(size: int, genome_length: int) -> Population:
    return [generate_genome(genome_length) for _ in range(size)]


def single_point_crossover(a: Genome, b: Genome) -> Tuple[Genome, Genome]:
    if len(a) != len(b):
        raise ValueError("Genomes a and b must be of same length")

    length = len(a)
    if length < 2:
        return a, b

    p = randint(1, length - 1)
    return a[0:p] + b[p:], b[0:p] + a[p:]


def mutation(genome: Genome, num: int = 1, probability: float = 0.5) -> Genome:
    for _ in range(num):
        index = randrange(len(genome))
        genome[index] = genome[index] if random() > probability else abs(genome[index] - 1)
    return genome


def population_fitness(population: Population, fitness_func: FitnessFunc) -> int:
    return sum([fitness_func(genome) for genome in population])


def selection_pair(population: Population, fitness_func: FitnessFunc) -> Population:
    return sample(
        population=generate_weighted_distribution(population, fitness_func),
        k=2
    )


def generate_weighted_distribution(population: Population, fitness_func: FitnessFunc) -> Population:
    result = []

    for gene in population:
        result += [gene] * int(fitness_func(gene)+1)

    return result


def sort_population(population: Population, fitness_func: FitnessFunc) -> Population:
    return sorted(population, key=fitness_func, reverse=True)


def genome_to_string(genome: Genome) -> str:
    return "".join(map(str, genome))


def print_stats(population: Population, generation_id: int, fitness_func: FitnessFunc):
    print("GENERATION %02d" % generation_id)
    print("=============")
    print("Population: [%s]" % ", ".join([genome_to_string(gene) for gene in population]))
    print("Avg. Fitness: %f" % (population_fitness(population, fitness_func) / len(population)))
    sorted_population = sort_population(population, fitness_func)
    print(
        "Best: %s (%f)" % (genome_to_string(sorted_population[0]), fitness_func(sorted_population[0])))
    print("Worst: %s (%f)" % (genome_to_string(sorted_population[-1]),
                              fitness_func(sorted_population[-1])))
    print("")

    return sorted_population[0]


def run_evolution(
        populate_func: PopulateFunc,
        fitness_func: FitnessFunc,
        fitness_limit: int,
        selection_func: SelectionFunc = selection_pair,
        crossover_func: CrossoverFunc = single_point_crossover,
        mutation_func: MutationFunc = mutation,
        generation_limit: int = 100,
        printer: Optional[PrinterFunc] = None) \
        -> Tuple[Population, int]:
    population = populate_func()

    i = 0
    for i in range(generation_limit):
        population = sorted(population, key=lambda genome: fitness_func(genome), reverse=True)

        if printer is not None:
            printer(population, i, fitness_func)

        if fitness_func(population[0]) >= fitness_limit:
            break

        next_generation = population[0:2]

        for j in range(int(len(population) / 2) - 1):
            parents = selection_func(population, fitness_func)
            offspring_a, offspring_b = crossover_func(parents[0], parents[1])
            offspring_a = mutation_func(offspring_a)
            offspring_b = mutation_func(offspring_b)
            next_generation += [offspring_a, offspring_b]

        population = next_generation

    return population, i